#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Relevation Password Printer
a command line interface to Revelation Password Manager.

Copyright (c) 2011, Toni Corvera
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

---------------------------------------------------------------------

Code based on Revelation's former BTS (no longer online, not archived?):
  (ref1) code:
  http://oss.wired-networks.net/bugzilla/attachment.cgi?id=13&action=view
  (ref2) bug report:
  http://oss.wired-networks.net/bugzilla/show_bug.cgi?id=111
    -> http://web.archive.org/http://oss.wired-networks.net/bugzilla/show_bug.cgi?id=111
(ref3) http://docs.python.org/library/zlib.html
"""
from Crypto.Cipher import AES
import ConfigParser
import getopt
import libxml2
import os
import stat
import sys
import zlib

__author__ = 'Toni Corvera'
__date__ = '$Date: 2011-06-30 19:56:51 +0200 (dj, 30 jun 2011) $'
__revision__ = '$Rev: 2016 $'
__version_info__ = ( 1, 0 ) #, 0 )
__version__ = '.'.join(map(str, __version_info__))
RELEASE=True

def usage(channel):
    ' Print help message '
    def p(s):
        channel.write(s)
    p('%s {-f passwordfile} {-p password | -0} [search] [search2] [...]\n' % sys.argv[0])
    p('\nOptions:\n')
    p('  -f FILE, --file=FILE         Revelation password file.\n')
    p('  -p PASS, --password=PASS     Master password.\n')
    p('  -s SEARCH, --search=SEARCH   Search for string.\n')
    p('  -i, --case-insensitive       Case insensitive search (default).\n')
    p('  -c, --case-sensitive         Case sensitive search.\n')
    p('  -a, --ask                    Interactively ask for password.\n')
    p('                               Note it will be displayed in clear as you\n')
    p('                               type it.\n')
    p('  -t TYPE, --type=TYPE         Print only entries of type TYPE.\n')
    p('                               With no search string, prints all entries of\n')
    p('                               type TYPE.\n')
    p('  -x, --xml                    Dump unencrypted XML document.\n')
    p('  -0, --stdin                  Read password from standard input.\n')
    p('  -h, --help                   Print help (this message).\n')
    p('  --version                    Print the program\'s version information.\n')
    p('\n')

def get_entries(document, xpath):
    """
    get_entries(xmlDoc, str xpath expression) -> list of xmlNode's

    Get entry nodes that match xpath
    """
    ctx = document.xpathNewContext()
    try:
        res = ctx.xpathEval(xpath)
    except libxml2.xpathError:
        if not RELEASE:
            sys.stderr.write('Failed with xpath expression: %s\n' % xpath)
        raise
    finally:
        ctx.xpathFreeContext()
    return res

def make_xpath_query(search_text=None, type_filter=None, ignore_case=True, negate_filter=False):
    '''
    make_xpath_query(str, str, bool, bool) -> str
    '''
    xpath = '/revelationdata//entry'
    if type_filter:
        sign = '='
        if negate_filter:
            sign = '!='
        xpath = '%s[@type%s"%s"]' % ( xpath, sign, type_filter )
    if search_text:
        xpath = xpath + '//text()'
        if ignore_case:
            xpath = '%s[contains(translate(., "ABCDEFGHIJKLMNOPQRSTUVWXYZ", "abcdefghijklmnopqrstuvwxyz"), "%s")]/../..' % ( xpath, search_text )
        else:
            xpath = '%s[contains(., "%s")]/../..' % ( xpath, search_text )
    return xpath

def dump_all_entries(xmldata):
    ' Dump all entries from xmldata, with no filter at all '
    doc = libxml2.parseDoc(xmldata)
    res = get_entries(doc, '//entry')
    nr = dump_result(res, 'all')
    doc.freeDoc()
    return nr

def dump_entries(xmldata, search_text=None, type_filter=None, ignore_case=True, negate_filter=False):
    ' Dump entries from xmldata that match criteria '
    doc = libxml2.parseDoc(xmldata)
    xpath = make_xpath_query(search_text, type_filter, ignore_case, negate_filter)
    res = get_entries(doc, xpath)
    query_desc = ''
    if search_text:
        query_desc = '"%s"' % search_text
    if type_filter:
        neg = ''
        if negate_filter:
            neg = 'not '
        if search_text:
            query_desc = '%s (\'%s%s\' entries)' % ( query_desc, neg, type_filter )
        else:
            query_desc = '%s%s entries' % ( neg, type_filter )
    nr = dump_result(res, query_desc)
    doc.freeDoc()
    return nr

def dump_result(res, query_desc):
    ''' Print query results.
    dump_result(list of entries, query description) -> int
    
    Note the XML document can't be freed before calling this function.
    '''
    print '-> Search %s: ' % query_desc,
    if not len(res):
        print 'No results'
        return False
    print '%d matches' % len(res)
    tagnames ={ 'generic-url': 'Url:',
            'generic-username': 'Username:',
            'generic-password': 'Password:',
            'generic-email': 'Email:',
            'generic-hostname': 'Hostname:',
            'generic-location': 'Location:',
            'generic-code': 'Code:',
            'generic-certificate': 'Certificate:',
            'generic-database': 'Database:',
            'generic-domain': 'Domain:',
            'generic-keyfile': 'Key file:',
            'generic-pin': 'PIN',
            'generic-port': 'Port'
                }
    for x in res:
        sys.stderr.write('-------------------------------------------------------------------------------\n')
        print ''
        for attr in x.properties: # Is it accessible directly?
            if attr.name == 'type':
                print 'Type:',attr.children
        for chld in x.children:
            n = chld.name
            val = chld.content
            if n == 'name':
                print 'Name:',val
            elif n == 'description':
                print 'Description:',val
            elif n == 'field':
                for attr in chld.properties:
                    if attr.name == 'id':
                        idv = attr.content
                        if idv in tagnames:
                            idv = tagnames[idv]
                        print idv,chld.content
        print ''
        # / for chld in x.children
    nr = len(res)
    plural = ''
    if nr > 1:
        plural = 's'
    sys.stderr.write('-------------------------------------------------------------------------------\n')
    sys.stderr.write('<- (end of %d result%s for {%s})\n\n' % ( nr, plural, query_desc ))
    return nr

def world_readable(path):
    ' Check if a file is readable by everyone '
    assert os.path.exists(path)
    if sys.platform == 'win32':
        return True
    st = os.stat(path)
    return bool(st.st_mode & stat.S_IROTH)

def load_config():
    """
    load_config() -> ( str file, str pass )
    Load configuration file is one is found
    """
    cfg = os.path.join(os.path.expanduser('~'), '.relevation.conf')
    pw = None
    fl = None
    if os.path.isfile(cfg):
        if os.access(cfg, os.R_OK):
            wr = world_readable(cfg)
            if wr:
                sys.stderr.write('Configuration (~/.relevation.conf) is world-readable!!!\n')
            parser = ConfigParser.ConfigParser()
            parser.read(cfg)
            ops = parser.options('relevation')
            if 'file' in ops:
                fl = os.path.expanduser(parser.get('relevation', 'file'))
            if 'password' in ops:
                if wr and sys.platform != 'win32': # TODO: how to check in windows?
                    sys.stderr.write('Your password can be read by anyone!!!\n')
                pw = parser.get('relevation', 'password')
        else: # exists but not readable
            sys.stderr.write('Configuration file (~/.relevation.conf) is not readable!\n')
    return ( fl, pw )

def main():
    datafile = None
    password = None
    # values to search for
    needles = []
    caseInsensitive = True
    # individual search: ( 'value to search', 'type of search', 'type of entry to filter' )
    searchTypes = []
    dump_xml = False

    sys.stderr.write('Relevation v%s, (c) 2011 Toni Corvera\n\n' % __version__)

    # ---------- OPTIONS ---------- #
    ( datafile, password ) = load_config()
    try:
        # gnu_getopt requires py >= 2.3
        ops, args = getopt.gnu_getopt(sys.argv[1:], 'f:p:s:0ciaht:x',
                        [ 'file=', 'password=', 'search=', 'stdin',
                          'case-sensitive', 'case-insensitive', 'ask',
                          'help', 'version', 'type=', 'xml' ])
    except getopt.GetoptError, err:
        print str(err)
        usage(sys.stderr)
        sys.exit(os.EX_USAGE)
    if args:
        needles = args
    
    if ( '-h', '' ) in ops or ( '--help', '' ) in ops:
        usage(sys.stdout)
        sys.exit(os.EX_OK)
    if ( '--version', '' ) in ops:
        release=''
        if not RELEASE:
            release=' [DEBUG]'
        print 'Relevation version %s%s' % ( __version__, release )
        print 'Python version %s' % sys.version
        sys.exit(os.EX_OK)
    
    for opt, arg in ops:
        if opt in ( '-f', '--file' ):
            datafile = arg
        elif opt in ( '-p', '--password' ):
            password = arg
        elif opt in ( '-a', '--ask', '-0', '--stdin' ):
            if opt in ( '-a', '--ask' ):
                sys.stderr.write('File password: ')
            password = sys.stdin.readline()
            password = password[:-1]
        elif opt in ( '-s', '--search' ):
            needles.append(arg)
        elif opt in ( '-i', '--case-insensitive' ):
            caseInsensitive = True
        elif opt in ( '-c', '--case-sensitive' ):
            caseInsensitive = False
        elif opt in ( '-t', '--type' ):
            iarg = arg.lower()
            neg = False
            if iarg.startswith('-'):
                iarg = iarg[1:]
                neg = True
            if not iarg in ( 'creditcard', 'cryptokey', 'database', 'door', 'email',
                            'folder', 'ftp', 'generic', 'phone', 'shell', 'website' ):
                sys.stderr.write('Warning: Type "%s" is not known by relevation.\n' % arg)
            searchTypes.append( ( iarg, neg ) )
        elif opt in ( '-x', '--xml' ):
            dump_xml = True
        else:
            sys.stderr.write('Unhandled option: %s\n' % opt)
            assert False, "internal error parsing options"
    if not datafile or not password:
        usage(sys.stderr)
        if not datafile:
            sys.stderr.write('Input password filename is required\n')
        if not password:
            sys.stderr.write('Password is required\n')
        sys.exit(os.EX_USAGE)
    
    # ---------- PASSWORDS FILE DECRYPTION ---------- #
    f = None
    try:
        if not os.access(datafile, os.R_OK):
            raise IOError('File \'%s\' not accessible' % datafile)
        f = open(datafile, "rb")
        # Encrypted data
        data = f.read()
    finally:
        if f:
            f.close()
    # Pad password
    password += (chr(0) * (32 - len(password)))
    # Data IV
    c = AES.new(password)
    iv = c.decrypt(data[12:28])
    # Decrypt. Decrypted data is compressed
    c = AES.new(password, AES.MODE_CBC, iv)
    cleardata_gz = c.decrypt(data[28:])
    # Length of data padding
    padlen = ord(cleardata_gz[-1])
    # Decompress actual data (15 is wbits [ref3] DON'T CHANGE, 2**15 is the (initial) buf size)
    xmldata = zlib.decompress(cleardata_gz[:-padlen], 15, 2**15)
    
    # ---------- QUERIES ---------- #
    if dump_xml:
        print xmldata
        sys.exit(os.EX_OK)
    # Multiply values to search by type of searches
    numhits = 0

    if not ( needles or searchTypes ): # No search nor filters, print all
        numhits = dump_all_entries(xmldata)
    elif not searchTypes: # Simple case, all searches are text searches
        for text in needles:
            numhits += dump_entries(xmldata, text, 'folder', caseInsensitive, True)
    elif needles: # Do a search filtered for each type
        for text in needles:
            for ( sfilter, negate ) in searchTypes:
                numhits += dump_entries(xmldata, text, sfilter, caseInsensitive,
                                negate_filter=negate)
    else: # Do a search only of types
        for ( sfilter, negate ) in searchTypes:
            numhits += dump_entries(xmldata, None, sfilter, negate_filter=negate)
    if numhits == 0:
        sys.exit(80)

if __name__ == '__main__':
    try:
        main()
    except libxml2.parserError as e:
        sys.stderr.write('XML parsing error\n')
        if not RELEASE:
            traceback.print_exc()
        sys.exit(os.EX_DATAERR)
    except IOError as e:
        if not RELEASE:
            traceback.print_exc()
        sys.stderr.write(str(e)+"\n")
        sys.exit(os.EX_IOERR)

# vim:set ts=4 et ai fileencoding=utf-8: #
